from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal

import nibabel as nib
import nibabel.spatialimages
import numpy as np

from medio.metadata.affine import Affine
from medio.metadata.metadata import MetaData

if TYPE_CHECKING:
    import os

    from numpy.typing import NDArray


def _reorient_affine(
    nib_affine: NDArray[np.floating],
    orig_shape: tuple[int, ...],
    desired_nib_axcodes: str | tuple[str, ...],
) -> tuple[NDArray[np.floating], tuple[int, ...]]:
    """Reorient a nibabel affine and shape to `desired_nib_axcodes` without loading pixel data.

    :param nib_affine: 4x4 nibabel-convention affine matrix
    :param orig_shape: spatial shape in the current orientation (x, y, z)
    :param desired_nib_axcodes: destination orientation, e.g. 'LPI' or ('L', 'P', 'I')
    :return: (new_nib_affine, new_shape)
    """
    if isinstance(desired_nib_axcodes, str):
        desired_nib_axcodes = tuple(desired_nib_axcodes)
    start_ornt = nib.orientations.io_orientation(nib_affine)
    end_ornt = nib.orientations.axcodes2ornt(desired_nib_axcodes)
    ornt_transform = nib.orientations.ornt_transform(start_ornt, end_ornt)
    new_affine = nib.orientations.inv_ornt_aff(ornt_transform, orig_shape) @ nib_affine
    new_shape = tuple(orig_shape[int(t[0])] for t in ornt_transform)
    return new_affine, new_shape


NibImage = nibabel.spatialimages.SpatialImage


class NibIO:
    coord_sys: ClassVar[Literal["nib"]] = "nib"
    RGB_DTYPE = np.dtype([("R", np.uint8), ("G", np.uint8), ("B", np.uint8)])
    RGBA_DTYPE = np.dtype([("R", np.uint8), ("G", np.uint8), ("B", np.uint8), ("A", np.uint8)])

    @staticmethod
    def read_img(
        input_path: str | os.PathLike[str],
        desired_axcodes: tuple[str, ...] | str | None = None,
        header: bool = False,
        channels_axis: int | None = None,
    ) -> tuple[NDArray[np.floating], MetaData[object]]:
        """
        Reads a NIFTI file and returns the image array and metadata
        :param input_path: path-like (str or pathlib.Path) of the image file to read
        :param desired_axcodes: str, tuple of str or None - the desired orientation of the image to be returned
        :param header: whether to include a header attribute with additional metadata in the returned metadata
        :param channels_axis: if not None and the array dtype is structured, stacks the channels along channels_axis
        :return: image array and corresponding metadata
        """
        img_struct = nib.load(input_path)
        orig_ornt_str = "".join(nib.aff2axcodes(img_struct.affine))
        if desired_axcodes is not None:
            img_struct = NibIO.reorient(img_struct, desired_axcodes)
        img = np.asanyarray(img_struct.dataobj)
        if channels_axis is not None:
            img = NibIO.unravel_array(img, channels_axis)
        affine = Affine(img_struct.affine)
        metadata = MetaData(affine=affine, orig_ornt=orig_ornt_str, coord_sys=NibIO.coord_sys)
        if header:
            metadata.header = {key: img_struct.header[key] for key in img_struct.header}
        return img, metadata

    @staticmethod
    def read_meta(
        input_path: str | os.PathLike[str],
        desired_axcodes: tuple[str, ...] | str | None = None,
        header: bool = False,
    ) -> MetaData[object]:
        """
        Read only the metadata (affine, orientation, spatial shape) of a NIfTI file without loading pixel data.
        :param input_path: path to the NIfTI file
        :param desired_axcodes: optional desired orientation, e.g. 'RAS'
        :param header: if True, populate metadata.header with NIfTI header fields
        :return: MetaData with spatial_shape set
        """
        # nibabel already does lazy loading of the pixel data, so we can read the affine and header
        # without loading the whole image.
        img_struct = nib.load(input_path)
        orig_ornt_str = "".join(nib.aff2axcodes(img_struct.affine))
        nib_affine: NDArray[np.floating] = img_struct.affine
        orig_shape: tuple[int, ...] = tuple(img_struct.shape[:3])

        if desired_axcodes is not None:
            new_affine, new_shape = _reorient_affine(nib_affine, orig_shape, desired_axcodes)
        else:
            new_affine, new_shape = nib_affine, orig_shape

        metadata: MetaData[object] = MetaData(
            affine=Affine(new_affine),
            orig_ornt=orig_ornt_str,
            coord_sys=NibIO.coord_sys,
            spatial_shape=new_shape,
        )
        metadata.spatial_shape = new_shape
        if header:
            metadata.header = {key: img_struct.header[key] for key in img_struct.header}
        return metadata

    @staticmethod
    def save_img(
        filename: str | os.PathLike[str],
        img: NDArray[np.floating],
        metadata: MetaData[object],
        use_original_ornt: bool = True,
        channels_axis: int | None = None,
    ) -> None:
        """
        Saves the given image as a NIFTI file.
        :param filename: path-like output filename, including a '.nii.gz' or '.nii' suffix
        :param img: image data array
        :param metadata: the matching metadata
        :param use_original_ornt: whether to use the original orientation of the image of not
        :param channels_axis: if not None gives the channels axis of img (for channeled images RGB/RGBA)
        """
        if channels_axis is not None:
            img = NibIO.pack_channeled_img(img, channels_axis)
        orig_coord_sys = metadata.coord_sys
        metadata.convert(NibIO.coord_sys)
        img_struct = nib.Nifti1Image(img, metadata.affine)
        desired_axcodes = metadata.orig_ornt if use_original_ornt else None
        metadata.convert(orig_coord_sys)
        img_struct = NibIO.reorient(img_struct, desired_axcodes)
        nib.save(img_struct, filename)

    @staticmethod
    def reorient(img_struct: NibImage, desired_axcodes: tuple[str, ...] | str | None) -> NibImage:
        """Reorient a nibabel image to a desired orientation described by desired_axcodes strings tuple, for example
        ('L', 'P', 'I'). If desired_axcodes is None it returns the given img_struct"""
        if desired_axcodes is not None:
            if isinstance(desired_axcodes, str):
                desired_axcodes = tuple(desired_axcodes)
            start_ornt = nib.orientations.io_orientation(img_struct.affine)
            end_ornt = nib.orientations.axcodes2ornt(desired_axcodes)
            ornt_tform = nib.orientations.ornt_transform(start_ornt, end_ornt)
            img_struct = img_struct.as_reoriented(ornt_tform)
        return img_struct

    @staticmethod
    def unravel_array(array: NDArray[np.generic], channels_axis: int = -1) -> NDArray[np.generic]:
        """Simplify array dtype if it is a structured data type. For example, if the array if of RGB dtype:
        np.dtype([('R', 'uint8'), ('G', 'uint8'), ('B', 'uint8')])
        Convert it into an array with dtype 'uint8' and 3 channels for RGB in an additional last dimension"""
        dtype = array.dtype
        if not (hasattr(dtype, "__len__") and len(dtype) > 1):
            return array
        return np.stack([array[field] for field in dtype.fields], axis=channels_axis)

    @staticmethod
    def pack_channeled_img(img: NDArray[np.uint8], channels_axis: int) -> NDArray[np.void]:
        dtype = img.dtype
        if not np.issubdtype(dtype, np.uint8):
            raise ValueError(f'RGB or RGBA images must have dtype "np.uint8", got: "{dtype}"')
        n_channels = img.shape[channels_axis]
        img = np.moveaxis(img, channels_axis, -1)
        r_channel = img[..., 0]
        if n_channels == 3:
            img_rgb = np.empty_like(r_channel, dtype=NibIO.RGB_DTYPE)
            img_rgb["R"] = r_channel
            img_rgb["G"] = img[..., 1]
            img_rgb["B"] = img[..., 2]
            return img_rgb
        elif n_channels == 4:
            img_rgba = np.empty_like(r_channel, dtype=NibIO.RGBA_DTYPE)
            img_rgba["R"] = r_channel
            img_rgba["G"] = img[..., 1]
            img_rgba["B"] = img[..., 2]
            img_rgba["A"] = img[..., 3]
            return img_rgba
        else:
            raise ValueError(f"Invalid number of channels: {n_channels}, should be 3 (RGB) or 4 (RGBA)")



from __future__ import annotations

import pprint
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generic, Literal, cast

import numpy as np
from nibabel import aff2axcodes

from medio.metadata.affine import Affine
from medio.metadata.convert_nib_itk import convert_affine, convert_nib_itk, inv_axcodes

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from typing_extensions import Self, TypeVar

    H = TypeVar("H", default=Any, covariant=True)
else:
    from typing import TypeVar

    H = TypeVar("H")

HeaderDict = dict[str, object]
CoordSys = Literal["itk", "nib"]


class MetaData(Generic[H]):
    affine: Affine
    orig_ornt: str | None
    coord_sys: CoordSys
    header: H | None
    spatial_shape: tuple[int, ...] | None
    _ornt: str | None

    def __init__(
        self,
        affine: Affine | NDArray[np.floating],
        orig_ornt: str | None = None,
        coord_sys: CoordSys = "itk",
        header: H | None = None,
        spatial_shape: tuple[int, ...] | None = None,
    ) -> None:
        """
        Initialize medical image's metadata
        :param affine: affine matrix of class Affine, numpy float array of shape (4, 4)
        :param orig_ornt: orientation string code, str of length 3 or None (was not computed because the image was not
        reoriented)
        :param coord_sys: 'itk' or 'nib', the coordinate system of the given affine and orientation: itk or nib (nifti)
        :param header: additional metadata dictionary with string keys
        """
        if not isinstance(affine, Affine):
            affine = Affine(affine)
        self.affine = affine
        self.orig_ornt = orig_ornt
        self._ornt = None
        self.coord_sys = self.check_valid_coord_sys(coord_sys)
        self.header = header
        self.spatial_shape = spatial_shape

    @staticmethod
    def check_valid_coord_sys(coord_sys: str) -> CoordSys:
        if coord_sys not in ("itk", "nib"):
            raise ValueError('Metadata coord_sys must be "itk" or "nib"')
        return cast("CoordSys", coord_sys)

    def __repr__(self) -> str:
        sep = " " if self.header is None else "\n"
        spatial_shape_str = f"\nSpatial shape: {self.spatial_shape}" if self.spatial_shape is not None else ""
        return (
            f"Affine:\n"
            f"{self.affine}\n"
            f"Spacing: {self.spacing}\n"
            f"Coordinate system: {self.coord_sys}\n"
            f"Orientation: {self.ornt}\n"
            f"Original orientation: {self.orig_ornt}"
            f"{spatial_shape_str}\n"
            f"Header:{sep}"
            f"{pprint.pformat(self.header, indent=4)}"
        )

    def convert(self, dest_coord_sys: CoordSys) -> None:
        """
        Converts the metadata coordinate system in-place to dest_coord_sys. Affects affine, ornt and orig_ornt
        :param dest_coord_sys: the destination coordinate system - 'itk' or 'nib' (nifti)
        """
        MetaData.check_valid_coord_sys(dest_coord_sys)  # runtime validation
        if dest_coord_sys != self.coord_sys:
            self.affine, self._ornt, self.orig_ornt = convert_nib_itk(self.affine, self._ornt, self.orig_ornt)
            self.coord_sys = dest_coord_sys

    def clone(self) -> Self:
        cloned = MetaData(
            affine=self.affine.clone(),
            orig_ornt=self.orig_ornt,
            coord_sys=self.coord_sys,
            header=deepcopy(self.header),
            spatial_shape=self.spatial_shape,
        )
        return cloned  # type: ignore[return-value]

    def get_ornt(self) -> str:
        """Returns current orientation based on the affine and coordinate system"""
        if self.coord_sys == "nib":
            return "".join(aff2axcodes(self.affine))
        else:  # "itk"
            return inv_axcodes("".join(aff2axcodes(convert_affine(self.affine))))

    @property
    def ornt(self) -> str:
        if self._ornt is None:
            self._ornt = self.get_ornt()
            # if self.orig_ornt is also None, the affine was not reoriented and the original orientation is the same
            if self.orig_ornt is None:
                self.orig_ornt = self._ornt
        return self._ornt

    @property
    def spacing(self) -> NDArray[np.floating]:
        return self.affine.spacing

    def is_right_handed_ornt(self) -> bool:
        """Check whether the affine orientation is right or left handed. The sign of the triple product of the
        direction matrix is calculated with a determinant. It should be +1 or -1 because it is a rotation matrix.
        +1 (-1) indicates right (left) handed orientation.
        To be used primarily before saving a dicom file or series"""
        if self.affine.dim != 3:
            raise ValueError("Right handed orientation is relevant only to a 3d space")
        return np.linalg.det(self.affine.direction) > 0


def is_right_handed_axcodes(axcodes: str) -> bool:
    if len(axcodes) == 2:
        return True
    if len(axcodes) != 3:
        raise ValueError(f'Invalid axcodes (not length 3 or 2): "{axcodes}"')
    letter_vec_dict = {
        "R": [1, 0, 0],
        "L": [-1, 0, 0],
        "A": [0, 1, 0],
        "P": [0, -1, 0],
        "I": [0, 0, 1],
        "S": [0, 0, -1],
    }
    u, v, n = [letter_vec_dict[letter] for letter in axcodes]
    ornt_sign = np.dot(np.cross(u, v), n)
    if ornt_sign not in (-1, 1):
        raise ValueError(f'Invalid axcodes: "{axcodes}"')
    return ornt_sign == 1


def flip_last_axcodes(axcodes: str) -> str:
    return axcodes[:-1] + inv_axcodes(axcodes[-1])


def check_dcm_ornt(desired_ornt: str | None, metadata: MetaData[Any], allow_dcm_reorient: bool = False) -> str:
    """Check whether the orientation desired_ornt is right handed before saving image as a dicom
    :param desired_ornt: the desired orientation for the saver
    :param metadata: if desired_ornt is None (not set), use metadata.ornt
    :param allow_dcm_reorient: whether to allow automatic reorientation to a right handed orientation or not
    :return: right handed desired_ornt or ValueError
    """
    # first set the desired orientation
    if desired_ornt is None:
        desired_ornt = metadata.ornt
    if is_right_handed_axcodes(desired_ornt):
        return desired_ornt
    else:
        right_handed_ornt = flip_last_axcodes(desired_ornt)
        if allow_dcm_reorient:
            return right_handed_ornt
        else:
            raise ValueError(
                f'The desired orientation "{desired_ornt}" is left handed, whereas saving dicom is '
                f"possible only with a right handed orientation. \nYou can either pass the saver "
                f"parameter allow_dcm_reorient=True to allow automatic reorientation (in this case to "
                f'"{right_handed_ornt}"), or \nreorient yourself before saving the image as a dicom.'
            )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Literal

import nibabel.spatialimages
import numpy as np
import pydicom
from dicom_numpy import combine_slices
from dicom_numpy.combine_slices import _extract_cosines, _validate_image_orientation

from medio.backends.nib_io import NibIO, _reorient_affine
from medio.backends.pdcm_unpack_ds import affine_from_dataset, unpack_dataset
from medio.metadata.convert_nib_itk import inv_axcodes
from medio.metadata.metadata import MetaData
from medio.metadata.pdcm_ds import MultiFrameFileDataset, convert_ds
from medio.utils.files import parse_series_uids

if TYPE_CHECKING:
    import os

    from numpy.typing import NDArray


class PdcmIO:
    coord_sys: ClassVar[Literal["itk"]] = "itk"
    # channels axes in the transposed image for pydicom and dicom-numpy. The actual axis is the first or the second
    # value of the tuple, according to the planar configuration (which is either 0 or 1)
    DEFAULT_CHANNELS_AXES_PYDICOM: tuple[int, int] = (0, -1)
    DEFAULT_CHANNELS_AXES_DICOM_NUMPY: tuple[int, int] = (0, 2)

    @staticmethod
    def read_img(
        input_path: str | os.PathLike[str],
        desired_ornt: str | None = None,
        header: bool = False,
        channels_axis: int | None = None,
        globber: str = "*",
        allow_default_affine: bool = False,
        series: str | int | None = None,
    ) -> tuple[NDArray[np.generic], MetaData[object]]:
        """
        Read a dicom file or folder (series) and return the numpy array and the corresponding metadata
        :param input_path: path-like object (str or pathlib.Path) of the file or directory to read
        :param desired_ornt: str, tuple of str or None - the desired orientation of the image to be returned
        :param header: whether to include a header attribute with additional metadata in the returned metadata (single
        file only)
        :param channels_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis in
        the returned image array
        :param globber: relevant for a directory - globber for selecting the series files (all files by default)
        :param allow_default_affine: whether to allow default affine when some tags are missing (multiframe file only)
        :param series: str or int of the series to read (in the case of multiple series in a directory)
        :return: numpy array and metadata
        """
        input_path = Path(input_path)
        # if there are channels, they must be in the last axis for the reorientation
        temp_channels_axis = -1
        if input_path.is_dir():
            img, metadata, channeled = PdcmIO.read_dcm_dir(
                input_path,
                header,
                globber,
                channels_axis=temp_channels_axis,
                series=series,
            )
        else:
            img, metadata, channeled = PdcmIO.read_dcm_file(
                input_path,
                header,
                allow_default_affine=allow_default_affine,
                channels_axis=temp_channels_axis,
            )
        img, metadata = PdcmIO.reorient(img, metadata, desired_ornt)
        # move the channels after the reorientation
        if channeled and channels_axis != temp_channels_axis:
            img = np.moveaxis(img, temp_channels_axis, channels_axis)
        return img, metadata

    @staticmethod
    def read_meta(
        input_path: str | os.PathLike[str],
        desired_ornt: str | None = None,
        header: bool = False,
        globber: str = "*",
        allow_default_affine: bool = False,
        series: str | int | None = None,
    ) -> MetaData[object]:
        """
        Read only the metadata (affine, orientation, spatial shape) of a DICOM file or directory without loading pixel
        data.
        :param input_path: path-like object (str or pathlib.Path) of the file or directory to read
        :param desired_ornt: optional orientation string to reorient the metadata, e.g. 'LPI'
        :param header: whether to include a header attribute (single file only; series raises NotImplementedError)
        :param globber: relevant for a directory - globber for selecting the series files
        :param allow_default_affine: use a default identity affine when geometric tags are missing (multiframe only)
        :param series: series to read when a directory has multiple series
        :return: MetaData with spatial_shape set
        """
        from medio.metadata.convert_nib_itk import convert_affine

        input_path = Path(input_path)
        if input_path.is_dir():
            slices = PdcmIO.extract_slices_no_pixels(input_path, globber, series)
            affine = PdcmIO._compute_series_affine(slices)
            ds0 = slices[0]
            spatial_shape: tuple[int, ...] = (int(ds0.Columns), int(ds0.Rows), len(slices))
            metadata: MetaData[object] = PdcmIO.aff2meta(affine)
            if header:
                raise NotImplementedError("header=True is currently not supported for a series")
        else:
            ds = pydicom.dcmread(input_path, stop_before_pixels=True)
            ds = convert_ds(ds)
            if ds.__class__ is MultiFrameFileDataset:
                affine = affine_from_dataset(ds, allow_default_affine=allow_default_affine)
                n_frames = int(ds.NumberOfFrames)
                spatial_shape = (int(ds.Columns), int(ds.Rows), n_frames)
            else:
                try:
                    _validate_image_orientation(ds.ImageOrientationPatient)
                    affine = PdcmIO._compute_single_slice_affine(ds)
                except AttributeError as e:
                    if allow_default_affine:
                        affine = np.eye(4, dtype=np.float32)
                    else:
                        raise AttributeError(str(e) + "\nTry using: allow_default_affine=True") from e
                spatial_shape = (int(ds.Columns), int(ds.Rows), 1)
            metadata = PdcmIO.aff2meta(affine)
            if header:
                metadata.header = {str(key): ds[key] for key in ds}

        if desired_ornt is not None and desired_ornt != metadata.ornt:
            orig_ornt = metadata.ornt
            # Convert affine to nib convention, reorient via pure matrix math, convert back
            nib_affine = convert_affine(metadata.affine)
            nib_desired = inv_axcodes(desired_ornt)
            new_nib_affine, spatial_shape = _reorient_affine(np.array(nib_affine), spatial_shape, nib_desired)
            metadata = MetaData(
                affine=convert_affine(new_nib_affine),
                orig_ornt=orig_ornt,
                coord_sys=PdcmIO.coord_sys,
                header=metadata.header,
            )

        metadata.spatial_shape = spatial_shape
        return metadata

    @staticmethod
    def _compute_single_slice_affine(ds: pydicom.Dataset) -> NDArray[np.float32]:
        """Compute affine for a regular (non-multiframe) single-slice DICOM dataset using header tags only."""
        row_cosine, column_cosine, _ = _extract_cosines(ds.ImageOrientationPatient)
        row_spacing, column_spacing = [float(x) for x in ds.PixelSpacing]
        slice_cosine = np.cross(row_cosine, column_cosine)
        slice_spacing = float(getattr(ds, "SpacingBetweenSlices", 1))
        transform = np.identity(4, dtype=np.float32)
        transform[:3, 0] = row_cosine * column_spacing
        transform[:3, 1] = column_cosine * row_spacing
        transform[:3, 2] = slice_cosine * slice_spacing
        transform[:3, 3] = [float(x) for x in ds.ImagePositionPatient]
        return transform

    @staticmethod
    def _compute_series_affine(slices: list[pydicom.Dataset]) -> NDArray[np.float32]:
        """Compute affine from a sorted list of DICOM slice datasets using header tags only."""
        ds0 = slices[0]
        row_cosine, column_cosine, _ = _extract_cosines(ds0.ImageOrientationPatient)
        row_spacing, column_spacing = [float(x) for x in ds0.PixelSpacing]
        first_pos = np.array([float(x) for x in ds0.ImagePositionPatient], dtype=np.float32)
        if len(slices) == 1:
            slice_cosine = np.cross(row_cosine, column_cosine)
            slice_spacing = float(getattr(ds0, "SpacingBetweenSlices", 1))
            slice_vector = slice_cosine * slice_spacing
        else:
            last_pos = np.array([float(x) for x in slices[-1].ImagePositionPatient], dtype=np.float32)
            slice_vector = (last_pos - first_pos) / (len(slices) - 1)
        transform = np.identity(4, dtype=np.float32)
        transform[:3, 0] = row_cosine * column_spacing
        transform[:3, 1] = column_cosine * row_spacing
        transform[:3, 2] = slice_vector
        transform[:3, 3] = first_pos
        return transform

    @staticmethod
    def read_dcm_file(
        filename: str | os.PathLike[str],
        header: bool = False,
        allow_default_affine: bool = False,
        channels_axis: int | None = None,
    ) -> tuple[NDArray[np.generic], MetaData[object], bool]:
        """
        Read a single dicom file.
        Return the image array, metadata, and whether it has channels
        """
        ds = pydicom.dcmread(filename)
        ds = convert_ds(ds)
        if ds.__class__ is MultiFrameFileDataset:
            img, affine = unpack_dataset(ds, allow_default_affine=allow_default_affine)
        else:
            img, affine = combine_slices([ds])
        metadata = PdcmIO.aff2meta(affine)
        if header:
            metadata.header = {str(key): ds[key] for key in ds}
        samples_per_pixel = ds.SamplesPerPixel
        img = PdcmIO.move_channels_axis(
            img,
            samples_per_pixel=samples_per_pixel,
            channels_axis=channels_axis,
            planar_configuration=ds.get("PlanarConfiguration", None),
            default_axes=PdcmIO.DEFAULT_CHANNELS_AXES_PYDICOM,
        )
        return img, metadata, samples_per_pixel > 1

    @staticmethod
    def read_dcm_dir(
        input_dir: str | os.PathLike[str],
        header: bool = False,
        globber: str = "*",
        channels_axis: int | None = None,
        series: str | int | None = None,
    ) -> tuple[NDArray[np.generic], MetaData[object], bool]:
        """
        Reads a 3D dicom image: input path can be a file or directory (DICOM series).
        Return the image array, metadata, and whether it has channels
        """
        # find all dicom files within the specified folder, read every file separately and sort them by InstanceNumber
        slices = PdcmIO.extract_slices(input_dir, globber=globber, series=series)
        img, affine = combine_slices(slices)
        metadata = PdcmIO.aff2meta(affine)
        if header:
            # TODO: add header support, something like
            #  metdata.header = [{str(key): ds[key] for key in ds.keys()} for ds in slices]
            raise NotImplementedError("header=True is currently not supported for a series")
        samples_per_pixel = slices[0].SamplesPerPixel
        img = PdcmIO.move_channels_axis(
            img,
            samples_per_pixel=samples_per_pixel,
            channels_axis=channels_axis,
            planar_configuration=slices[0].get("PlanarConfiguration", None),
            default_axes=PdcmIO.DEFAULT_CHANNELS_AXES_DICOM_NUMPY,
        )
        return img, metadata, samples_per_pixel > 1

    @staticmethod
    def extract_slices(
        input_dir: str | os.PathLike[str],
        globber: str = "*",
        series: str | int | None = None,
    ) -> list[pydicom.Dataset]:
        """Extract slices from input_dir and return them sorted"""
        files = Path(input_dir).glob(globber)
        slices = [pydicom.dcmread(filename) for filename in files]

        # filter by Series Instance UID
        datasets = {}
        for slc in slices:
            key = slc.SeriesInstanceUID
            datasets[key] = [*datasets.get(key, []), slc]

        series_uid = parse_series_uids(input_dir, datasets.keys(), series, globber)
        slices = datasets[series_uid]

        slices.sort(key=lambda ds: ds.get("InstanceNumber", 0))
        return slices

    @staticmethod
    def extract_slices_no_pixels(
        input_dir: str | os.PathLike[str],
        globber: str = "*",
        series: str | int | None = None,
    ) -> list[pydicom.Dataset]:
        """Extract slices from input_dir without loading pixel data (header-only).
        Returns sorted list of pydicom Datasets read with stop_before_pixels=True."""
        files = list(Path(input_dir).glob(globber))
        slices = [pydicom.dcmread(f, stop_before_pixels=True) for f in files]

        datasets: dict[str, list[pydicom.Dataset]] = {}
        for slc in slices:
            key = slc.SeriesInstanceUID
            datasets[key] = [*datasets.get(key, []), slc]

        series_uid = parse_series_uids(input_dir, datasets.keys(), series, globber)
        slices = datasets[series_uid]
        slices.sort(key=lambda ds: ds.get("InstanceNumber", 0))
        return slices

    @staticmethod
    def aff2meta(affine: NDArray[np.floating]) -> MetaData[object]:
        return MetaData(affine, coord_sys=PdcmIO.coord_sys)

    @staticmethod
    def move_channels_axis(
        array: NDArray[np.generic],
        samples_per_pixel: int,
        channels_axis: int | None = None,
        planar_configuration: int | None = None,
        default_axes: tuple[int, int] = DEFAULT_CHANNELS_AXES_PYDICOM,
    ) -> NDArray[np.generic]:
        """Move the channels axis from the original axis to the destined channels_axis"""
        if (samples_per_pixel == 1) or (channels_axis is None):
            # no rearrangement is needed
            return array

        # extract the original channels axis
        if planar_configuration not in [0, 1]:
            raise ValueError(f"Invalid Planar Configuration value: {planar_configuration}")

        orig_axis = default_axes[planar_configuration]
        flag = True  # original channels axis is assigned
        shape = array.shape
        # validate that the assigned axis matches samples_per_pixel, if not - try to search for it
        if shape[orig_axis] != samples_per_pixel:
            flag = False
            for i, sz in enumerate(shape):
                if sz == samples_per_pixel:
                    orig_axis = i
                    flag = True
                    break

        if not flag:
            raise ValueError("The original channels axis was not detected")

        return np.moveaxis(array, orig_axis, channels_axis)

    @staticmethod
    def reorient(
        img: NDArray[np.generic],
        metadata: MetaData[object],
        desired_ornt: str | None,
    ) -> tuple[NDArray[np.generic], MetaData[object]]:
        """
        Reorient img array and affine (in the metadata) to desired_ornt (str) using nibabel.
        desired_ornt is in itk convention.
        Note that if img has channels (RGB for example), they must be in last axis
        """
        if (desired_ornt is None) or (desired_ornt == metadata.ornt):
            return img, metadata
        # convert from pydicom (itk) to nibabel convention
        metadata.convert(NibIO.coord_sys)
        orig_ornt = metadata.ornt
        desired_ornt = inv_axcodes(desired_ornt)
        # use nibabel for the reorientation
        img_struct = nibabel.spatialimages.SpatialImage(img, metadata.affine)
        reoriented_img_struct = NibIO.reorient(img_struct, desired_ornt)

        img = np.asanyarray(reoriented_img_struct.dataobj)
        metadata = MetaData(
            reoriented_img_struct.affine,
            orig_ornt=orig_ornt,
            coord_sys=NibIO.coord_sys,
            header=metadata.header,
        )
        # convert back to pydicom convention
        metadata.convert(PdcmIO.coord_sys)
        return img, metadata

    @staticmethod
    def save_arr2dcm_file(
        output_filename: str | os.PathLike[str],
        template_filename: str | os.PathLike[str],
        img_arr: NDArray[np.generic],
        dtype: np.dtype[np.generic] | str | None = None,
        keep_rescale: bool = False,
    ) -> None:
        """
        Writes a dicom single file image using template file, without the intensity transformation from template dataset
        unless keep_rescale is True
        :param output_filename: path-like object of the output file to be saved
        :param template_filename: the single dicom scan whose metadata is used
        :param img_arr: numpy array of the image to be saved, should be in the same orientation as template_filename
        :param dtype: the dtype for the numpy array, for example 'int16'. If None - will use the dtype of the template
        :param keep_rescale: whether to keep intensity rescale values
        """
        ds = pydicom.dcmread(template_filename)
        ds = convert_ds(ds)
        if not keep_rescale:
            if isinstance(ds, MultiFrameFileDataset):
                ds.del_intensity_trans()
            else:
                del ds.RescaleSlope
                del ds.RescaleIntercept
        if dtype is None:
            img_arr = img_arr.astype(ds.pixel_array.dtype, copy=False)
        else:
            img_arr = img_arr.astype(dtype, copy=False)
        ds.PixelData = img_arr.tobytes()
        ds.save_as(output_filename)



from __future__ import annotations

from functools import partial

from pydicom.uid import generate_uid as _generate_uid

# Given by Medical Connections (http://www.medicalconnections.co.uk/FreeUID.html)
MEDIO_ROOT_UID = "1.2.826.0.1.3680043.10.513."


generate_uid = partial(_generate_uid, prefix=MEDIO_ROOT_UID)



"""
NiBabel <-> ITK orientation and affine conversion utilities.
The conventions of nibabel and itk are different and this module supplies functions which convert between
these conventions.

Orientation
-----------
In nibabel each axis code indicates the ending direction - RAS+: L -> R, P -> A, I -> S
In itk it corresponds to the converse of nibabel - RAS: R -> L, A -> P, S -> I

Affine
------
In itk, the direction matrix (3x3 upper left affine with unit spacings) of RAS orientation image is:
[[1, 0, 0],
 [0, 1, 0],
 [0, 0, -1]]

In nibabel it is LPI+:
[[-1, 0, 0],
 [0, -1, 0],
 [0, 0, -1]]

The matrix convert_aff_mat accounts for this difference (for all possible orientations, not only RAS).

Usage
=====
Works both ways: itk -> nib and nib -> itk, the usage is the same:
>>> new_affine, new_axcodes = convert_nib_itk(affine, axcodes)
"""

from __future__ import annotations

from typing import TYPE_CHECKING, overload

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeVar, TypeVarTuple, Unpack

from medio.metadata.affine import Affine
from medio.utils.two_way_dict import TwoWayDict

# store compactly axis directions codes
axes_inv = TwoWayDict()
axes_inv["R"] = "L"
axes_inv["A"] = "P"
axes_inv["S"] = "I"


@overload
def inv_axcodes(axcodes: str) -> str: ...


@overload
def inv_axcodes(axcodes: None) -> None: ...


@overload
def inv_axcodes(axcodes: str | None) -> str | None: ...


def inv_axcodes(axcodes: str | None) -> str | None:
    """Inverse axes codes chars, for example: SPL -> IAR"""
    if axcodes is None:
        return None
    new_axcodes = ""
    for code in axcodes:
        new_axcodes += axes_inv[code]
    return new_axcodes


AffineOrNdarray = TypeVar("AffineOrNdarray", Affine, NDArray[np.floating])


def convert_affine(affine: AffineOrNdarray) -> AffineOrNdarray:
    # conversion matrix of the affine from itk to nibabel and vice versa
    convert_aff_mat = np.diag([-1, -1, 1, 1])
    # for 2d image:
    if affine.shape[0] == 3:
        convert_aff_mat = np.diag([-1, -1, 1])
    new_affine = convert_aff_mat @ affine
    if isinstance(affine, Affine):
        return Affine(new_affine)
    return new_affine


VariadicAxCodes = TypeVarTuple("VariadicAxCodes")


if TYPE_CHECKING:
    import collections.abc as cx
    from typing import Any, Protocol

    F = TypeVar("F", bound=cx.Callable[..., Any])

    class _ConvertNibItkFunc(Protocol):
        def __call__(self, affine: AffineOrNdarray, *axcodes: str | None) -> Any: ...

    def _type_check_axcodes(f: F, /) -> F | _ConvertNibItkFunc:
        """Decorate a function to make it only accept variadic positional `axcodes` arguments of type `str` or `None`"""
        return f
else:

    def _type_check_axcodes(f, /):
        return f


@_type_check_axcodes
def convert_nib_itk(
    affine: AffineOrNdarray, *axcodes: Unpack[VariadicAxCodes]
) -> tuple[AffineOrNdarray, Unpack[VariadicAxCodes]]:
    """Convert affine and orientations (original and current orientations) from nibabel to itk and vice versa"""
    new_affine = convert_affine(affine)
    new_axcodes = []
    for axcode in axcodes:
        new_axcodes += [inv_axcodes(axcode)]

    return (new_affine, *new_axcodes)



from __future__ import annotations

import os
import pprint
from pathlib import Path
from typing import TYPE_CHECKING, Union

from typing_extensions import TypeGuard

if TYPE_CHECKING:
    from collections.abc import Iterable

PathLike = Union[os.PathLike[str], str]


def is_file_suffix(filename: PathLike, suffixes: tuple[str, ...], check_exist: bool = True) -> bool:
    """
    is_file + check for suffix
    :param filename: pathlike object
    :param suffixes: tuple of possible suffixes
    :param check_exist: whether to check the file's existence
    :return: bool
    """
    if check_exist and not Path(filename).is_file():
        return False
    return str(filename).endswith(suffixes)


def is_nifti(filename: PathLike, check_exist: bool = True) -> TypeGuard[PathLike]:
    return is_file_suffix(
        filename,
        (".nii.gz", ".nii", ".img.gz", ".img", ".hdr"),
        check_exist=check_exist,
    )


def is_dicom(filename: PathLike, check_exist: bool = True) -> TypeGuard[PathLike]:
    return is_file_suffix(filename, (".dcm", ".dicom", ".DCM", ".DICOM"), check_exist=check_exist)


def make_empty_dir(dir_path: PathLike, parents: bool = False) -> None:
    """Make an empty directory. If it exists - check that it is empty"""
    dir_path = Path(dir_path)
    try:
        dir_path.mkdir(parents=parents, exist_ok=False)
    except FileExistsError:
        # the directory exists
        try:
            next(dir_path.glob("*"))
        except StopIteration:
            pass  # the directory exists but empty - ok
        else:
            raise FileExistsError(f'The directory "{dir_path}" is not empty')


def make_dir(dir_path: PathLike, parents: bool = False, exist_ok: bool = False) -> None:
    if exist_ok:
        Path(dir_path).mkdir(parents=parents, exist_ok=exist_ok)
    else:
        make_empty_dir(dir_path, parents)


def parse_series_uids(
    input_dir: PathLike,
    series_uids: Iterable[str],
    series: str | int | None = None,
    globber: str | None = None,
) -> str:
    """Receive an input dir, an iterable of series UIDs, and a series (UID string or int),
    return a series uid according to series_uids and series"""
    keys = sorted(series_uids)
    num_series = len(keys)
    if num_series == 0:
        raise FileNotFoundError(
            f'No DICOMs in:\n"{input_dir}"' + (f'\nwith globber="{globber}"' if globber is not None else "")
        )

    if num_series == 1:
        return keys[0]

    # if there is more than a single series
    else:
        if series is None:
            raise ValueError(
                f'The directory: "{input_dir}"\n'
                "contains more than a single DICOM series. "
                "The following series were identified according to their Series Instance UID:\n"
                f"{pprint.pformat(keys)}\n"
                "Try passing: series=series_uid, where series_uid is a one of the strings above,\n"
                f"or an integer between 0 and {num_series - 1} corresponding to one of them."
            )
        elif isinstance(series, int):
            return keys[series]
        else:
            if series not in keys:
                raise ValueError(f"The series:\n'{series}'\nis not one of the following:\n{pprint.pformat(keys)}")
            return series



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import itk
import itk.support.types as itkt
import numpy as np

from medio.metadata.affine import Affine
from medio.metadata.dcm_uid import generate_uid
from medio.metadata.itk_orientation import itk_orientation_code
from medio.metadata.metadata import MetaData, check_dcm_ornt
from medio.utils.files import is_dicom, make_dir, parse_series_uids

if TYPE_CHECKING:
    import os

    from numpy.typing import NDArray

    from medio.metadata.metadata import CoordSys


class ItkIO:
    coord_sys: ClassVar[Literal["itk"]] = "itk"
    DEFAULT_COMPONENTS_AXIS: int = 0  # in the transposed image
    # default image type:
    dimension: int = 3
    pixel_type = itk.ctype("short")  # signed short - int16
    image_type = itk.Image[pixel_type, dimension]

    @staticmethod
    def read_img(
        input_path: str | os.PathLike[str],
        desired_axcodes: str | tuple[str, ...] | None = None,
        header: bool = False,
        components_axis: int | None = None,
        pixel_type: itkt.PixelTypes | None = pixel_type,
        fallback_only: bool = True,
        series: str | int | None = None,
        private_tags: bool = False,
    ) -> tuple[NDArray[np.generic], MetaData[object]]:
        """
        The main reader function, reads images and performs reorientation and unpacking
        :param input_path: path of image file or directory containing dicom series
        :param desired_axcodes: string or tuple - e.g. 'LPI', ('R', 'A', 'S')
        :param header: whether to include a header attribute with additional metadata in the returned metadata
        :param components_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis
        :param pixel_type: preferred itk pixel type for the image
        :param fallback_only: if True, finds the pixel_type automatically and uses pixel_type only if failed
        :param series: str or int of the series to read (in the case of multiple series in a directory)
        :return: numpy image and metadata object which includes pixdim, affine, original orientation string and
        coordinates system
        """
        if header:
            # Currently, only DICOM will use imageio (GDCM). For NIfTI, we'll use ITK default reader.
            imageio = itk.GDCMImageIO.New()
            if private_tags:
                imageio.LoadPrivateTagsOn()
        else:
            imageio = None
        input_path = Path(input_path)
        if input_path.is_dir():
            # We assume that the directory contains dicom series, do imageio will work, if used.
            if imageio is not None:
                # fallback_only=True would skip the imageio, so disable it
                fallback_only = False
            img = ItkIO.read_dir(str(input_path), pixel_type, fallback_only, series, imageio)
        elif input_path.is_file():
            # If the input file is not a dicom (e.g. NIfTI), fallback to not use imageio.
            use_dicom_imageio = imageio is not None and imageio.CanReadFile(str(input_path))
            if use_dicom_imageio:
                # fallback_only=True would skip the imageio, so disable it
                fallback_only = False
            else:
                imageio = None
            img = ItkIO.read_img_file(str(input_path), pixel_type, fallback_only, imageio)
        else:
            raise FileNotFoundError(f'No such file or directory: "{input_path}"')

        affine = ItkIO.get_img_aff(img)
        metadata = MetaData(affine=affine, coord_sys=ItkIO.coord_sys)
        if (desired_axcodes is None) or (desired_axcodes == metadata.ornt):
            image_np = ItkIO.itk_img_to_array(img)
        else:
            orig_ornt = metadata.ornt
            img, _ = ItkIO.reorient(img, desired_axcodes)
            image_np, affine = ItkIO.unpack_img(img)
            metadata = MetaData(affine=affine, orig_ornt=orig_ornt, coord_sys=ItkIO.coord_sys)
        if header:
            metadict = imageio.GetMetaDataDictionary() if imageio else img.GetMetaDataDictionary()
            metadata.header = {key: metadict[key] for key in metadict.GetKeys() if not key.startswith("ITK_")}

        # TODO: consider unifying with PdcmIO.move_channels_axis
        n_components = img.GetNumberOfComponentsPerPixel()
        if (n_components > 1) and (components_axis is not None):
            # assert image_np.shape[ItkIO.DEFAULT_COMPONENTS_AXIS] == n_components
            image_np = np.moveaxis(image_np, ItkIO.DEFAULT_COMPONENTS_AXIS, components_axis)

        return image_np, metadata

    @staticmethod
    def read_meta(
        input_path: str | os.PathLike[str],
        desired_axcodes: str | tuple[str, ...] | None = None,
        header: bool = False,
        pixel_type: itkt.PixelTypes | None = pixel_type,
        fallback_only: bool = True,
        series: str | int | None = None,
        private_tags: bool = False,
    ) -> MetaData[object]:
        """
        Read only the metadata (affine, orientation, spatial shape) without loading pixel data.
        Uses ITK's UpdateOutputInformation() which processes only the file header.
        :param input_path: path of image file or directory containing dicom series
        :param desired_axcodes: optional orientation string, e.g. 'LPI'
        :param header: whether to include a header attribute with additional metadata
        :param pixel_type: preferred itk pixel type
        :param fallback_only: if True, auto-detect pixel type
        :param series: series to read when a directory has multiple series
        :param private_tags: if True, also load private DICOM tags (requires header=True)
        :return: MetaData with spatial_shape set
        """

        from medio.backends.nib_io import _reorient_affine
        from medio.metadata.convert_nib_itk import convert_affine, inv_axcodes

        input_path = Path(input_path)
        image_type = itk.Image[pixel_type, ItkIO.dimension]

        if input_path.is_dir():
            filenames = ItkIO.extract_series(str(input_path), series)
            reader = itk.ImageSeriesReader[image_type].New()
            if isinstance(filenames, str):
                reader.SetFileNames([filenames])
            else:
                reader.SetFileNames(filenames)
            if header:
                imageio = itk.GDCMImageIO.New()
                if private_tags:
                    imageio.LoadPrivateTagsOn()
                reader.SetImageIO(imageio)
            else:
                imageio = None
            reader.UpdateOutputInformation()
            img = reader.GetOutput()
        elif input_path.is_file():
            reader = itk.ImageFileReader[image_type].New()
            reader.SetFileName(str(input_path))
            if header and is_dicom(str(input_path), check_exist=False):
                imageio = itk.GDCMImageIO.New()
                if private_tags:
                    imageio.LoadPrivateTagsOn()
                reader.SetImageIO(imageio)
            else:
                imageio = None
            reader.UpdateOutputInformation()
            img = reader.GetOutput()
        else:
            raise FileNotFoundError(f'No such file or directory: "{input_path}"')

        affine = ItkIO.get_img_aff(img)

        # Spatial shape in ITK is z, y, x from GetSize(); after the .T transpose in itk_img_to_array it becomes x, y, z
        itk_size = img.GetLargestPossibleRegion().GetSize()
        spatial_shape: tuple[int, ...] = tuple(itk_size[i] for i in range(ItkIO.dimension))

        metadata: MetaData[object] = MetaData(affine=affine, coord_sys=ItkIO.coord_sys, spatial_shape=spatial_shape)

        if desired_axcodes is not None and desired_axcodes != metadata.ornt:
            orig_ornt = metadata.ornt
            # Reorient using pure affine math (no pixel data)
            # Convert to nib convention for _reorient_affine
            nib_affine = convert_affine(affine)
            nib_desired = (
                inv_axcodes(desired_axcodes)
                if isinstance(desired_axcodes, str)
                else tuple(inv_axcodes(c) for c in desired_axcodes)
            )
            new_nib_affine, new_shape = _reorient_affine(nib_affine, spatial_shape, nib_desired)
            new_affine = convert_affine(new_nib_affine)
            metadata = MetaData(
                affine=new_affine, orig_ornt=orig_ornt, coord_sys=ItkIO.coord_sys, spatial_shape=new_shape
            )

        if header:
            try:
                metadict = imageio.GetMetaDataDictionary() if imageio else img.GetMetaDataDictionary()
                metadata.header = {key: metadict[key] for key in metadict.GetKeys() if not key.startswith("ITK_")}
            except Exception:
                metadata.header = {}

        return metadata

    @staticmethod
    def save_img(
        filename: str | os.PathLike[str],
        image_np: NDArray[np.generic],
        metadata: MetaData[object],
        use_original_ornt: bool = True,
        components_axis: int | None = None,
        allow_dcm_reorient: bool = False,
        compression: bool = False,
    ) -> None:
        """
        Save an image file with itk
        :param filename: the filename to save, str or os.PathLike
        :param image_np: the image's numpy array
        :param metadata: the corresponding metadata
        :param use_original_ornt: whether to save in the original orientation or not
        :param components_axis: if not None - the image has more than 1 component (e.g. RGB) and the components are in
        components_axis
        :param allow_dcm_reorient: whether to allow automatic reorientation to a right handed orientation or not
        :param compression: use compression or not
        """
        is_dcm = is_dicom(filename, check_exist=False)
        if is_dcm:
            image_np = ItkIO.prepare_dcm_array(image_np, is_vector=components_axis is not None)
        image = ItkIO.prepare_image(
            image_np,
            metadata,
            use_original_ornt,
            components_axis=components_axis,
            is_dcm=is_dcm,
            allow_dcm_reorient=allow_dcm_reorient,
        )
        ItkIO.save_img_file(image, str(filename), compression=compression)

    @staticmethod
    def prepare_image(
        image_np: NDArray[np.generic],
        metadata: MetaData[object],
        use_original_ornt: bool,
        components_axis: int | None = None,
        is_dcm: bool = False,
        allow_dcm_reorient: bool = False,
    ) -> object:
        """Prepare image for saving"""
        orig_coord_sys = metadata.coord_sys
        metadata.convert(ItkIO.coord_sys)
        desired_ornt = metadata.orig_ornt if use_original_ornt else None
        if is_dcm:
            # checking right-handed orientation before saving a dicom file/series
            desired_ornt = check_dcm_ornt(desired_ornt, metadata, allow_dcm_reorient=allow_dcm_reorient)
        image = ItkIO.pack2img(image_np, metadata.affine, components_axis=components_axis)
        if (desired_ornt is not None) and (desired_ornt != metadata.ornt):
            image, _ = ItkIO.reorient(image, desired_ornt)
        metadata.convert(orig_coord_sys)
        return image

    @staticmethod
    def prepare_dcm_array(image_np: NDArray[np.generic], is_vector: bool = False) -> NDArray[np.generic]:
        """Change image_np to correct data type for saving a single dicom file"""
        if is_vector:
            dcm_dtypes = [np.uint8]
        else:
            # for 3d image the supported data types are:
            dcm_dtypes = [np.uint8, np.uint16]
            # if the image is 2d it can be signed
            if np.squeeze(image_np).ndim == 2:
                dcm_dtypes = [np.int16, *dcm_dtypes]

        if image_np.dtype in dcm_dtypes:
            return image_np

        for dtype in dcm_dtypes:
            arr = image_np.astype(dtype, copy=False)
            if np.array_equal(arr, image_np):
                return arr

        raise NotImplementedError(
            "Saving a single dicom file with ItkIO is currently supported only for \n"
            "1. 2d images - int16, uint16, uint8\n"
            "2. 3d images with integer nonnegative values - uint8, uint16\n"
            "3. 2d/3d RGB[A] images - uint8 (with channels_axis)\n"
            "For negative values, try to save a dicom directory or use PdcmIO.save_arr2dcm_file"
        )

    @staticmethod
    def read_img_file(
        filename: str,
        pixel_type: itkt.PixelTypes | None = None,
        fallback_only: bool = False,
        imageio: object | None = None,
    ) -> object:
        """Common pixel types: itk.SS (int16), itk.US (uint16), itk.UC (uint8)"""
        return itk.imread(filename, pixel_type, fallback_only, imageio)

    @staticmethod
    def read_img_file_long(filename: str, image_type: type = image_type) -> tuple[object, str]:
        """Longer version of itk.imread that returns the itk image and io engine string"""
        reader = itk.ImageFileReader[image_type].New()
        reader.LoadPrivateTagsOn()
        reader.SetFileName(filename)
        reader.Update()
        image_io = str(reader.GetImageIO()).split(" ")[0]
        image = reader.GetOutput()
        return image, image_io

    @staticmethod
    def save_img_file(image: object, filename: str, compression: bool = False) -> None:
        itk.imwrite(image, filename, compression)

    @staticmethod
    def save_img_file_long(image: object, filename: str, compression: bool = False) -> None:
        image_type = type(image)
        writer = itk.ImageFileWriter[image_type].New()
        if compression:
            writer.UseCompressionOn()
        writer.UseInputMetaDataDictionaryOn()
        writer.SetFileName(filename)
        writer.SetInput(image)
        writer.Update()

    @staticmethod
    def itk_img_to_array(img_itk: object) -> NDArray[np.generic]:
        """
        Swap the axes to the usual x, y, z convention in RAI orientation
        (originally z, y, x)
        """
        # the transpose here is equivalent to keep_axes=True
        img_array = itk.array_from_image(img_itk).T
        return img_array

    @staticmethod
    def array_to_itk_img(img_array: NDArray[np.generic], components_axis: int | None = None) -> object:
        """Set components_axis to not None for vector images, e.g. RGB"""
        is_vector = False
        if components_axis is not None:
            img_array = np.moveaxis(img_array, components_axis, ItkIO.DEFAULT_COMPONENTS_AXIS)
            is_vector = True
        # copy is crucial for the ordering
        img_itk = itk.image_from_array(img_array.T.copy(), is_vector=is_vector)
        return img_itk

    @staticmethod
    def unpack_img(img: object) -> tuple[NDArray[np.generic], Affine]:
        image_np = ItkIO.itk_img_to_array(img)
        affine = ItkIO.get_img_aff(img)
        return image_np, affine

    @staticmethod
    def from_itk_img(
        itk_img: object,
        desired_ornt: str | None = None,
        coord_sys: CoordSys | None = "itk",
        components_axis: int | None = None,
    ) -> tuple[NDArray[np.generic], MetaData[object]]:
        """
        Convert an ITK image object to a medio (numpy array, MetaData) pair.
        :param itk_img: an itk.Image object
        :param desired_ornt: optional orientation for reorientation in the convention of `coord_sys`.
            When coord_sys is None, must be in ITK convention.
        :param coord_sys: convention of desired_ornt and returned MetaData ('itk', 'nib', or None).
            None means no conversion is applied (ITK convention is returned).
        :param components_axis: for multi-component images (e.g. RGB), move components to this axis
        :return: (np_image, metadata) — same format as read_img
        """
        from medio.metadata.convert_nib_itk import inv_axcodes

        itk_desired_ornt: str | None = desired_ornt
        if (coord_sys is not None) and (coord_sys != ItkIO.coord_sys):
            itk_desired_ornt = inv_axcodes(desired_ornt)

        affine = ItkIO.get_img_aff(itk_img)
        metadata: MetaData[object] = MetaData(affine=affine, coord_sys=ItkIO.coord_sys)
        if (itk_desired_ornt is None) or (itk_desired_ornt == metadata.ornt):
            img = itk_img
            image_np = ItkIO.itk_img_to_array(img)
        else:
            orig_ornt = metadata.ornt
            img, _ = ItkIO.reorient(itk_img, itk_desired_ornt)
            image_np, affine = ItkIO.unpack_img(img)
            metadata = MetaData(affine=affine, orig_ornt=orig_ornt, coord_sys=ItkIO.coord_sys)

        n_components = img.GetNumberOfComponentsPerPixel()
        if (n_components > 1) and (components_axis is not None):
            image_np = np.moveaxis(image_np, ItkIO.DEFAULT_COMPONENTS_AXIS, components_axis)

        if coord_sys is not None:
            metadata.convert(coord_sys)
        return image_np, metadata

    @staticmethod
    def to_itk_img(
        np_image: NDArray[np.generic],
        metadata: MetaData[object],
        components_axis: int | None = None,
    ) -> object:
        """
        Convert a medio (numpy array, MetaData) pair to an ITK image object.
        Handles any coord_sys automatically; the caller's MetaData is never mutated.
        :param np_image: the image numpy array
        :param metadata: the MetaData object (any coord_sys)
        :param components_axis: axis in np_image holding image components (e.g. RGB channels)
        :return: itk.Image with spatial metadata applied
        """
        metadata_itk = metadata.clone()
        metadata_itk.convert(ItkIO.coord_sys)
        return ItkIO.pack2img(np_image, metadata_itk.affine, components_axis=components_axis)

    @staticmethod
    def get_img_aff(img: object) -> Affine:
        direction = itk.array_from_vnl_matrix(img.GetDirection().GetVnlMatrix().as_matrix())
        spacing = itk.array_from_vnl_vector(img.GetSpacing().GetVnlVector())
        origin = itk.array_from_vnl_vector(img.GetOrigin().GetVnlVector())
        return Affine(direction=direction, spacing=spacing, origin=origin)

    @staticmethod
    def pack2img(
        image_np: NDArray[np.generic],
        affine: Affine | NDArray[np.floating],
        components_axis: int | None = None,
    ) -> object:
        image = ItkIO.array_to_itk_img(image_np, components_axis)
        ItkIO.set_img_aff(image, affine)
        return image

    @staticmethod
    def set_img_aff(image: object, affine: Affine | NDArray[np.floating]) -> None:
        if not isinstance(affine, Affine):
            affine = Affine(affine)

        dimension = image.GetImageDimension()
        direction_arr, spacing, origin = affine.direction, affine.spacing, affine.origin

        # setting metadata
        spacing_vec = itk.Vector[itk.D, dimension]()
        spacing_vec.SetVnlVector(itk.vnl_vector_from_array(spacing.astype("float")))
        image.SetSpacing(spacing_vec)

        image.SetOrigin(origin.astype("float"))

        direction_mat = itk.vnl_matrix_from_array(direction_arr.astype("float"))
        direction = itk.Matrix[itk.D, dimension, dimension](direction_mat)
        image.SetDirection(direction)

    @staticmethod
    def reorient(img: object, desired_orientation: int | tuple[str, ...] | str | None) -> tuple[object, int | None]:
        if desired_orientation is None:
            return img, None

        image_type = type(img)
        orient = itk.OrientImageFilter[image_type, image_type].New()
        orient.UseImageDirectionOn()
        orient.SetInput(img)
        if isinstance(desired_orientation, (str, tuple)):
            desired_orientation = itk_orientation_code(desired_orientation)
        orient.SetDesiredCoordinateOrientation(desired_orientation)
        orient.Update()
        reoriented_itk_img = orient.GetOutput()
        original_orientation_code = orient.GetGivenCoordinateOrientation()
        return reoriented_itk_img, original_orientation_code

    @staticmethod
    def read_dir(
        dirname: str,
        pixel_type: itk.itkCType | None = None,
        fallback_only: bool = False,
        series: str | int | None = None,
        imageio: object | None = None,
    ) -> object:
        """
        Read a dicom directory. If there is more than one series in the directory an error is raised
        (unless the series argument is used properly).
        Shorter option for a single series (provided the slices order is known):
        >>> itk.imread([filename0, filename1, ...])
        """
        filenames = ItkIO.extract_series(dirname, series)
        return itk.imread(filenames, pixel_type, fallback_only, imageio)

    @staticmethod
    def extract_series(dirname: str, series: str | int | None = None) -> list[str] | str:
        """Extract series filenames from the directory dirname"""
        names_generator = itk.GDCMSeriesFileNames.New()
        names_generator.SetDirectory(dirname)

        series_uids = names_generator.GetSeriesUIDs()
        series_uid = parse_series_uids(dirname, series_uids, series)

        filenames = names_generator.GetFileNames(series_uid)
        if len(filenames) == 1:
            filenames = filenames[0]  # there is a single image in the series

        return filenames

    @staticmethod
    def save_dcm_dir(
        dirname: str | os.PathLike[str],
        image_np: NDArray[np.generic],
        metadata: MetaData[object],
        use_original_ornt: bool = True,
        components_axis: int | None = None,
        parents: bool = False,
        exist_ok: bool = False,
        allow_dcm_reorient: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Save a 3d numpy array image_np as a dicom series of 2d dicom slices in the directory dirname
        :param dirname: the directory to save in the files, str or pathlib.Path. If it exists - must be empty
        :param image_np: the image's numpy array
        :param metadata: the corresponding metadata
        :param use_original_ornt: whether to save in the original orientation or not
        :param components_axis: if not None - the image has more than 1 component (e.g. RGB) and the components are in
        components_axis
        :param parents: if True, creates also the parents of dirname
        :param exist_ok: if True, non-empty existing directory will not raise an error
        :param allow_dcm_reorient: whether to allow automatic reorientation to a right-handed orientation or not
        :param kwargs: optional kwargs passed to ItkIO.dcm_metadata: pattern, metadata_dict
        """
        image = ItkIO.prepare_image(
            image_np,
            metadata,
            use_original_ornt,
            components_axis=components_axis,
            is_dcm=True,
            allow_dcm_reorient=allow_dcm_reorient,
        )
        image_type = type(image)
        _, (pixel_type, _) = itk.template(image)
        image2d_type = itk.Image[pixel_type, 2]
        writer = itk.ImageSeriesWriter[image_type, image2d_type].New()
        make_dir(dirname, parents, exist_ok)
        # Generate necessary metadata and filenames per slice:
        mdict_list, filenames = ItkIO.dcm_series_metadata(image, dirname, **kwargs)
        metadict_vec = itk.vector[itk.MetaDataDictionary](mdict_list)
        writer.SetMetaDataDictionaryArray(metadict_vec)
        writer.SetFileNames(filenames)
        dicom_io = itk.GDCMImageIO.New()
        dicom_io.KeepOriginalUIDOn()
        writer.SetImageIO(dicom_io)
        writer.SetInput(image)
        writer.Update()

    @staticmethod
    def dcm_series_metadata(
        image: object,
        dirname: str | os.PathLike[str],
        pattern: str = "IM{}.dcm",
        metadata_dict: dict[str, str] | None = None,
    ) -> tuple[list[object], list[str]]:
        """
        Return dicom series metadata per slice and filenames
        :param image: the full itk image to be saved as dicom series
        :param dirname: the directory name
        :param pattern: str pattern for the filenames to save, including a placeholder ('{}') for the slice number
        :param metadata_dict: dictionary of metadata for adding tags or overriding the default values. For example,
        metadata_dict={'0008|0060': 'US'} will override the default 'CT' modality and set it to 'US' (ultrasound)
        :return: metadata dictionaries per slice, slice filenames
        """
        # The number of slices
        n = image.GetLargestPossibleRegion().GetSize().GetElement(2)

        # Shared properties for all the n slices:
        mdict = itk.MetaDataDictionary()

        # Series Instance UID
        mdict["0020|000e"] = generate_uid()
        # Study Instance UID
        mdict["0020|000d"] = generate_uid()

        date, time = datetime.now().strftime("%Y%m%d %H%M%S.%f").split()
        # Study Date
        mdict["0008|0020"] = date
        # Series Date
        mdict["0008|0021"] = date
        # Content Date
        mdict["0008|0023"] = date
        # Study Time
        mdict["0008|0030"] = time
        # Series Time
        mdict["0008|0031"] = time

        # Pixel Spacing (not necessary - automatically saved)
        spacing = image.GetSpacing()
        mdict["0028|0030"] = f"{spacing[0]}\\{spacing[1]}"
        # Spacing Between Slices
        mdict["0018|0088"] = str(spacing[2])
        # Image Orientation (Patient)
        orientation_str = "\\".join(
            [str(image.GetDirection().GetVnlMatrix().get(i, j)) for j in range(2) for i in range(3)]
        )
        mdict["0020|0037"] = orientation_str
        # Patient Position
        mdict["0018|5100"] = ""

        # Number of Frames
        mdict["0028|0008"] = "1"
        # Number of Slices
        mdict["0054|0081"] = str(n)
        # Modality
        mdict["0008|0060"] = "CT"

        if metadata_dict is not None:
            for key, val in metadata_dict.items():
                mdict[key] = val

        # Per slice properties:
        mdict_list = []
        filenames = []

        for i in range(n):
            # copy the shared properties dict:
            mdict_i = itk.MetaDataDictionary(mdict)
            # Instance Number
            mdict_i["0020|0013"] = str(i + 1)
            # Image Position (Patient)
            position = image.TransformIndexToPhysicalPoint([0, 0, i])
            position_str = "\\".join([str(position[i]) for i in range(3)])
            mdict_i["0020|0032"] = position_str

            mdict_list += [mdict_i]
            filenames += [str(Path(dirname) / pattern.format(i + 1))]

        return mdict_list, filenames



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

from medio.metadata.affine import Affine
from medio.metadata.metadata import MetaData
from medio.read_save import read_img, save_img
from medio.utils.explicit_slicing import explicit_inds

if TYPE_CHECKING:
    import os

    import numpy as np
    from numpy.typing import NDArray
    from typing_extensions import Self


class MedImg:
    np_image: NDArray[np.generic]
    metadata: MetaData[Any]

    def __init__(
        self,
        np_image: NDArray[np.generic] | None,
        metadata: MetaData[Any] | None,
        filename: str | os.PathLike[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Class for a single medical image, represented by numpy array and metadata object - image affine, original
        orientation and coordinate system. The class allows performing operations on the image with respective update of
        the metadata (mainly affine).
        The class can be instantiated also with a filename:
        >>> MedImg(None, None, '../scan1.mhd')
        With possible kwargs for the reader.
        """
        if filename is not None:
            # warn that using filename is deprecated and will be removed in future, and suggest using from_file instead
            warnings.warn(
                "Instantiating MedImg with a filename is deprecated and will be removed in a future release. "
                "Please use the from_file class method instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            np_image, metadata = read_img(filename, **kwargs)
        self.np_image = np_image
        self.metadata = metadata  # type: ignore[assignment]

    @classmethod
    def from_file(cls, filename: str | os.PathLike[str], **kwargs: Any) -> MedImg:
        """Alternative constructor to create MedImg directly from a file."""
        np_image, metadata = read_img(filename, **kwargs)
        return cls(np_image, metadata)

    def save(self, filename: str | os.PathLike[str], **kwargs: Any) -> None:
        save_img(filename, self.np_image, self.metadata, **kwargs)

    def __getitem__(self, item: tuple[object, ...]) -> Self:
        """
        This method allows cropping and basic down-sampling:
        >>> mimg = MedImg(np_image, metadata)
        >>> new_mimg = mimg[:, 4:-4, ::3]
        >>> print(new_mimg.metadata)
        Ellipsis (...) is also supported
        """
        np_image = self.np_image[item]
        start, _stop, stride = explicit_inds(item, self.np_image.shape)
        affine = Affine(self.metadata.affine.copy())
        affine.origin = affine.index2coord(start)
        affine.spacing = affine.spacing * stride
        metadata = MetaData(affine, self.metadata.orig_ornt, self.metadata.coord_sys)
        return MedImg(np_image, metadata)  # type: ignore[return-value]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, overload

from medio.backends.itk_io import ItkIO
from medio.backends.nib_io import NibIO
from medio.backends.pdcm_io import PdcmIO
from medio.metadata.convert_nib_itk import inv_axcodes
from medio.utils.files import is_nifti

if TYPE_CHECKING:
    import os

    import numpy as np
    from numpy.typing import NDArray

    from medio.metadata.metadata import CoordSys, HeaderDict, MetaData

ReadBackend = Literal["itk", "nib", "pdcm", "pydicom"]
WriteBackend = Literal["itk", "nib"]


@overload
def read_img(
    input_path: str | os.PathLike[str],
    desired_ornt: str | None = ...,
    backend: ReadBackend | None = ...,
    dtype: np.dtype[np.generic] | type | None = ...,
    *,
    header: Literal[True],
    channels_axis: int | None = ...,
    coord_sys: CoordSys | None = ...,
    **kwargs: Any,
) -> tuple[NDArray[np.generic], MetaData[HeaderDict]]: ...


@overload
def read_img(
    input_path: str | os.PathLike[str],
    desired_ornt: str | None = ...,
    backend: ReadBackend | None = ...,
    dtype: np.dtype[np.generic] | type | None = ...,
    header: bool = ...,
    channels_axis: int | None = ...,
    coord_sys: CoordSys | None = ...,
    **kwargs: Any,
) -> tuple[NDArray[np.generic], MetaData[object]]: ...


def read_img(
    input_path: str | os.PathLike[str],
    desired_ornt: str | None = None,
    backend: ReadBackend | None = None,
    dtype: np.dtype[np.generic] | type | None = None,
    header: bool = False,
    channels_axis: int | None = -1,
    coord_sys: CoordSys | None = "itk",
    **kwargs: Any,
) -> tuple[NDArray[np.generic], MetaData[object] | MetaData[HeaderDict]]:
    """
    Read medical image with nibabel or itk
    :param input_path: str or os.PathLike, the input path of image file or a directory containing dicom series
    :param desired_ornt: optional parameter for reorienting the image to a desired orientation, e.g. 'RAS'.
    The desired_ornt string is in the convention of `coord_sys` argument (itk by default).
    :param backend: optional parameter for setting the reader backend: 'itk', 'nib', 'pdcm' (also 'pydicom') or None
    :param dtype: equivalent to np_image.astype(dtype) if dtype is not None
    :param header: if True, the returned metadata will include a header attribute with additional metadata dictionary as
    read by the backend. Note: currently, this is supported for files only
    :param channels_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis in the
    returned image array
    :param coord_sys: the coordinate system (or convention) of the `desired_ornt` parameter and the returned metadata.
    It can be 'itk', 'nib' or None, and is 'itk' by default.
    None means that the backend will determine coord_sys, but it can lead to a backend-dependent array and metadata
    :return: numpy image and metadata object
    """
    nib_reader_data = (NibIO.read_img, NibIO.coord_sys)
    itk_reader_data = (ItkIO.read_img, ItkIO.coord_sys)
    pdcm_reader_data = (PdcmIO.read_img, PdcmIO.coord_sys)
    if backend is None:
        if is_nifti(input_path):
            reader, reader_sys = nib_reader_data
        else:
            reader, reader_sys = itk_reader_data
    else:
        if backend == "nib":
            reader, reader_sys = nib_reader_data
        elif backend == "itk":
            reader, reader_sys = itk_reader_data
        elif backend in ("pdcm", "pydicom"):
            reader, reader_sys = pdcm_reader_data
        else:
            raise ValueError('The backend argument must be one of: "itk", "nib", "pdcm" (or "pydicom"), None')

    if (coord_sys is not None) and (coord_sys != reader_sys):
        desired_ornt = inv_axcodes(desired_ornt)

    np_image, metadata = reader(input_path, desired_ornt, header, channels_axis, **kwargs)

    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    if coord_sys is not None:
        metadata.convert(coord_sys)
    return np_image, metadata


def read_meta(
    input_path: str | os.PathLike[str],
    desired_ornt: str | None = None,
    backend: ReadBackend | None = None,
    header: bool = False,
    coord_sys: CoordSys | None = "itk",
    **kwargs: Any,
) -> MetaData[object]:
    """
    Read only the metadata of a medical image (affine, orientation, spatial shape) without loading pixel data.
    :param input_path: str or os.PathLike, the input path of an image file or a DICOM directory
    :param desired_ornt: optional orientation to reorient the returned metadata, e.g. 'RAS'. Uses the `coord_sys`
    convention (itk by default).
    :param backend: optional backend to use: 'itk', 'nib', 'pdcm' (also 'pydicom') or None (auto-detected)
    :param header: if True, the returned metadata will include a header attribute with the raw backend metadata
    dictionary. Currently supported for single files only (not DICOM series).
    :param coord_sys: coordinate system of `desired_ornt` and of the returned metadata: 'itk', 'nib' or None.
    :return: MetaData object with spatial_shape set to the image dimensions
    """
    nib_reader_data = (NibIO.read_meta, NibIO.coord_sys)
    itk_reader_data = (ItkIO.read_meta, ItkIO.coord_sys)
    pdcm_reader_data = (PdcmIO.read_meta, PdcmIO.coord_sys)
    if backend is None:
        if is_nifti(input_path):
            reader_meta, reader_sys = nib_reader_data
        else:
            reader_meta, reader_sys = itk_reader_data
    else:
        if backend == "nib":
            reader_meta, reader_sys = nib_reader_data
        elif backend == "itk":
            reader_meta, reader_sys = itk_reader_data
        elif backend in ("pdcm", "pydicom"):
            reader_meta, reader_sys = pdcm_reader_data
        else:
            raise ValueError('The backend argument must be one of: "itk", "nib", "pdcm" (or "pydicom"), None')

    if (coord_sys is not None) and (coord_sys != reader_sys):
        desired_ornt = inv_axcodes(desired_ornt)

    metadata = reader_meta(input_path, desired_ornt, header, **kwargs)

    if coord_sys is not None:
        metadata.convert(coord_sys)
    return metadata


def save_img(
    filename: str | os.PathLike[str],
    np_image: NDArray[np.generic],
    metadata: MetaData[Any],
    use_original_ornt: bool = True,
    backend: WriteBackend | None = None,
    dtype: np.dtype[np.generic] | type | None = None,
    channels_axis: int | None = None,
    mkdir: bool = False,
    parents: bool = False,
    **kwargs: Any,
) -> None:
    """
    Save numpy image with corresponding metadata to file
    :param filename: str or os.PathLike, the output filename
    :param np_image: the numpy image
    :param metadata: the metadata of the image
    :param use_original_ornt: whether to save in the original orientation stored in metadata.orig_ornt or not
    :param backend: optional - 'itk', 'nib' or None
    :param dtype: equivalent to np_image.astype(dtype) if dtype is not None
    :param channels_axis: if not None - the image is channeled (e.g. RGB) and the channels are in channels_axis
    :param mkdir: if True, creates the directory of `filename`
    :param parents: to be used with `mkdir=True`. If True, creates also the parent directories
    """
    nib_writer = NibIO.save_img
    itk_writer = ItkIO.save_img
    if backend is None:
        writer = nib_writer if is_nifti(filename, check_exist=False) else itk_writer
    else:
        if backend == "nib":
            writer = nib_writer
        elif backend == "itk":
            writer = itk_writer
        else:
            raise ValueError('The backend argument must be one of: "itk", "nib", None')
    if mkdir:
        Path(filename).parent.mkdir(parents=parents, exist_ok=True)
    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    writer(filename, np_image, metadata, use_original_ornt, channels_axis, **kwargs)


def save_dir(
    dirname: str | os.PathLike[str],
    np_image: NDArray[np.generic],
    metadata: MetaData[Any],
    use_original_ornt: bool = True,
    dtype: np.dtype[np.generic] | type | None = None,
    channels_axis: int | None = None,
    parents: bool = False,
    exist_ok: bool = False,
    allow_dcm_reorient: bool = False,
    **kwargs: Any,
) -> None:
    """
    Save image as a dicom directory. See medio.backends.itk_io.ItkIO.save_dcm_dir documentation.
    dtype is equivalent to passing image_np.astype(dtype) if dtype is not None
    """
    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    ItkIO.save_dcm_dir(
        dirname, np_image, metadata, use_original_ornt, channels_axis, parents, exist_ok, allow_dcm_reorient, **kwargs
    )



from __future__ import annotations

from typing import Generic, TypeVar, Union

KT = TypeVar("KT")
VT = TypeVar("VT")


class TwoWayDict(dict[Union[KT, VT], Union[KT, VT]], Generic[KT, VT]):
    """Dictionary which contains key-value + value-key pairs: {key: value, value: key}"""

    def __setitem__(self, key: KT | VT, value: KT | VT) -> None:
        # Remove any previous connections with these values
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        dict.__setitem__(self, key, value)
        dict.__setitem__(self, value, key)

    def __delitem__(self, key: KT | VT) -> None:
        dict.__delitem__(self, self[key])
        dict.__delitem__(self, key)

    def __len__(self) -> int:
        """Returns the number of connections"""
        return dict.__len__(self) // 2



from __future__ import annotations

from importlib.metadata import version

from medio.backends.itk_io import ItkIO
from medio.medimg import MedImg
from medio.metadata.affine import Affine
from medio.metadata.metadata import CoordSys, MetaData
from medio.read_save import read_img, read_meta, save_dir, save_img

__version__ = version("medio")

__all__ = [
    "Affine",
    "CoordSys",
    "ItkIO",
    "MedImg",
    "MetaData",
    "__version__",
    "read_img",
    "read_meta",
    "save_dir",
    "save_img",
]
